"""revamp_keypairs

Revision ID: 4b8a66fb8d82
Revises: 854bd902b1bc
Create Date: 2017-09-13 01:57:42.355633

"""
from alembic import op

# revision identifiers, used by Alembic.
revision = '4b8a66fb8d82'
down_revision = '854bd902b1bc'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.alter_column('keypairs', column_name='billing_plan', new_column_name='resource_policy')
    op.create_index(op.f('ix_keypairs_is_active'), 'keypairs', ['is_active'], unique=False)
    op.create_index(op.f('ix_keypairs_user_id'), 'keypairs', ['user_id'], unique=False)
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.alter_column('keypairs', column_name='resource_policy', new_column_name='billing_plan')
    op.drop_index(op.f('ix_keypairs_user_id'), table_name='keypairs')
    op.drop_index(op.f('ix_keypairs_is_active'), table_name='keypairs')
    # ### end Alembic commands ###



"""add_startsat_field_into_kernels

Revision ID: da24ff520049
Revises: 529113b08c2c
Create Date: 2020-06-18 20:47:22.152831

"""
from alembic import op
import sqlalchemy as sa
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = 'da24ff520049'
down_revision = '529113b08c2c'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column('kernels', sa.Column('starts_at', sa.DateTime(timezone=True), nullable=True))
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column('kernels', 'starts_at')
    # ### end Alembic commands ###



from __future__ import annotations

import asyncio
from contextvars import ContextVar
from datetime import datetime
import itertools
import logging
import pkg_resources
from typing import (
    Any,
    Awaitable,
    Dict,
    List,
    Mapping,
    MutableMapping,
    Sequence,
    Tuple,
    Union,
    TYPE_CHECKING,
)

import aioredis
import aioredlock
from dateutil.tz import tzutc
import sqlalchemy as sa
from sqlalchemy.ext.asyncio import AsyncConnection as SAConnection
from sqlalchemy.sql.expression import true

from ai.backend.common.docker import ImageRef
from ai.backend.common.events import (
    AgentStartedEvent,
    DoScheduleEvent,
    SessionCancelledEvent,
    SessionEnqueuedEvent,
    SessionScheduledEvent,
    SessionTerminatedEvent,
    EventDispatcher,
    EventProducer,
)
from ai.backend.common.logging import BraceStyleAdapter
from ai.backend.common.types import (
    aobject,
    AgentId,
    ClusterMode,
    ResourceSlot,
    SessionId,
)

from ..api.exceptions import InstanceNotAvailable
from ..distributed import GlobalTimer
from ..defs import REDIS_STREAM_DB
from ..exceptions import convert_to_status_data
from ..models import (
    agents, kernels, keypairs, scaling_groups,
    recalc_agent_resource_occupancy,
    AgentStatus, KernelStatus,
    AGENT_RESOURCE_OCCUPYING_KERNEL_STATUSES,
)
from ..models.utils import sql_json_increment, sql_json_merge
from . import (
    PredicateResult,
    PendingSession,
    ExistingSession,
    SchedulingContext,
    AgentContext,
    AgentAllocationContext,
    AbstractScheduler,
    KernelInfo,
    KernelAgentBinding,
)
from .predicates import (
    check_reserved_batch_session,
    check_concurrency,
    check_dependencies,
    check_keypair_resource_limit,
    check_group_resource_limit,
    check_domain_resource_limit,
    check_scaling_group,
)

if TYPE_CHECKING:
    from ..config import LocalConfig, SharedConfig
    from ..registry import AgentRegistry

__all__ = (
    'load_scheduler',
    'SchedulerDispatcher',
)

log = BraceStyleAdapter(logging.getLogger('ai.backend.manager.scheduler'))

_log_fmt: ContextVar[str] = ContextVar('_log_fmt')
_log_args: ContextVar[Tuple[Any, ...]] = ContextVar('_log_args')


def load_scheduler(name: str, scheduler_configs: Mapping[str, Any]) -> AbstractScheduler:
    entry_prefix = 'backendai_scheduler_v10'
    for entrypoint in pkg_resources.iter_entry_points(entry_prefix):
        if entrypoint.name == name:
            log.debug('loading scheduler plugin "{}" from {}', name, entrypoint.module_name)
            scheduler_cls = entrypoint.load()
            scheduler_config = scheduler_configs.get(name, {})
            return scheduler_cls(scheduler_config)
    raise ImportError('Cannot load the scheduler plugin', name)


def merge_resource(src: MutableMapping[str, Any], val: MutableMapping[str, Any]) -> None:
    for k in val.keys():
        if k in src.keys():
            src[k] += val[k]
        else:
            src[k] = val[k]


StartTaskArgs = Tuple[
    Tuple[Any, ...],
    SchedulingContext,
    Tuple[PendingSession, List[KernelAgentBinding]],
    List[Tuple[str, Union[Exception, PredicateResult]]],
]


class SchedulerDispatcher(aobject):

    config: LocalConfig
    shared_config: SharedConfig
    registry: AgentRegistry

    lock_manager: aioredlock.Aioredlock
    timer_redis: aioredis.Redis
    event_dispatcher: EventDispatcher
    event_producer: EventProducer
    schedule_timer: GlobalTimer

    def __init__(
        self,
        local_config: LocalConfig,
        shared_config: SharedConfig,
        event_dispatcher: EventDispatcher,
        event_producer: EventProducer,
        registry: AgentRegistry,
    ) -> None:
        self.local_config = local_config
        self.shared_config = shared_config
        self.event_dispatcher = event_dispatcher
        self.event_producer = event_producer
        self.registry = registry
        self.db = registry.db
        self.schedule_lock_timeout = 120.0

    async def __ainit__(self) -> None:
        log.info('Session scheduler started')
        self.registry.event_dispatcher.consume(SessionEnqueuedEvent, None, self.schedule)
        self.registry.event_dispatcher.consume(SessionTerminatedEvent, None, self.schedule)
        self.registry.event_dispatcher.consume(AgentStartedEvent, None, self.schedule)
        self.registry.event_dispatcher.consume(DoScheduleEvent, None, self.schedule)
        redis_url = self.shared_config.get_redis_url(db=REDIS_STREAM_DB)
        self.lock_manager = aioredlock.Aioredlock(
            [str(redis_url)],
            # we have explicit semantics: temporary locking failure -> try at the next chance,
            # such as scheduling timer ticks or upon scheduling events
            retry_count=2,
        )
        self.timer_redis = await aioredis.create_redis(str(redis_url))
        self.schedule_timer = GlobalTimer(
            self.timer_redis,
            "scheduler_tick",
            self.event_producer,
            lambda: DoScheduleEvent(),
            interval=10.0,
        )
        await self.schedule_timer.join()

    async def close(self) -> None:
        log.info('Session scheduler stopped')
        await self.schedule_timer.leave()
        self.timer_redis.close()
        await self.timer_redis.wait_closed()

    async def schedule(
        self,
        context: None,
        source: AgentId,
        event: SessionEnqueuedEvent | SessionTerminatedEvent | AgentStartedEvent | DoScheduleEvent,
    ) -> None:
        try:
            lock = await self.lock_manager.lock(
                'manager.scheduler',
                lock_timeout=self.schedule_lock_timeout,
            )
            async with lock:
                await self.schedule_impl()
        except aioredlock.LockError:
            log.debug('schedule(): temporary locking failure; will be retried.')
            # The dispatcher will try the next chance.

    async def schedule_impl(self) -> None:
        log.debug('schedule(): triggered')
        known_slot_types = await self.shared_config.get_resource_slots()
        sched_ctx = SchedulingContext(
            registry=self.registry,
            known_slot_types=known_slot_types,
        )
        start_task_args: List[StartTaskArgs] = []

        # We use short transaction blocks to prevent deadlock timeouts under heavy loads
        # because this scheduling handler will be executed by only one process.
        # It is executed under a globally exclusive context using aioredlock.
        async with self.db.connect() as agent_db_conn, \
                   self.db.connect() as kernel_db_conn:
            async with agent_db_conn.begin():
                query = (
                    sa.select([agents.c.scaling_group])
                    .select_from(agents)
                    .where(agents.c.status == AgentStatus.ALIVE)
                    .group_by(agents.c.scaling_group)
                )
                result = await agent_db_conn.execute(query)
                schedulable_scaling_groups = [
                    row.scaling_group for row in result.fetchall()
                ]
            for sgroup_name in schedulable_scaling_groups:
                try:
                    args_list = await self._schedule_in_sgroup(
                        sched_ctx, agent_db_conn, kernel_db_conn, sgroup_name,
                    )
                    start_task_args.extend(args_list)
                except InstanceNotAvailable:
                    # Proceed to the next scaling group and come back later.
                    log.debug('schedule({}): instance not available', sgroup_name)
                except Exception as e:
                    log.exception('schedule({}): scheduling error!\n{}', sgroup_name, repr(e))

        # At this point, all scheduling decisions are made
        # and the resource occupation is committed to the database.
        if start_task_args:
            start_coros = []
            for args in start_task_args:
                start_coros.append(self.start_session(*args))
            await asyncio.gather(*start_coros)

    async def _load_scheduler(
        self,
        db_conn: SAConnection,
        sgroup_name: str,
    ) -> AbstractScheduler:
        query = (
            sa.select([scaling_groups.c.scheduler])
            .select_from(scaling_groups)
            .where(scaling_groups.c.name == sgroup_name)
        )
        result = await db_conn.execute(query)
        scheduler_name = result.scalar()
        return load_scheduler(scheduler_name, self.shared_config['plugins']['scheduler'])

    async def _schedule_in_sgroup(
        self,
        sched_ctx: SchedulingContext,
        agent_db_conn: SAConnection,
        kernel_db_conn: SAConnection,
        sgroup_name: str,
    ) -> List[StartTaskArgs]:
        async with kernel_db_conn.begin():
            scheduler = await self._load_scheduler(kernel_db_conn, sgroup_name)
            pending_sessions = await _list_pending_sessions(kernel_db_conn, sgroup_name)
            existing_sessions = await _list_existing_sessions(kernel_db_conn, sgroup_name)
        log.debug('running scheduler (sgroup:{}, pending:{}, existing:{})',
                  sgroup_name, len(pending_sessions), len(existing_sessions))
        zero = ResourceSlot()
        args_list: List[StartTaskArgs] = []
        while len(pending_sessions) > 0:
            async with agent_db_conn.begin():
                candidate_agents = await _list_agents_by_sgroup(agent_db_conn, sgroup_name)
                total_capacity = sum((ag.available_slots for ag in candidate_agents), zero)

            picked_session_id = scheduler.pick_session(
                total_capacity,
                pending_sessions,
                existing_sessions,
            )
            if picked_session_id is None:
                # no session is picked.
                # continue to next sgroup.
                return []
            for picked_idx, sess_ctx in enumerate(pending_sessions):
                if sess_ctx.session_id == picked_session_id:
                    break
            else:
                # no matching entry for picked session?
                raise RuntimeError('should not reach here')
            sess_ctx = pending_sessions.pop(picked_idx)

            log_fmt = 'schedule(s:{}, type:{}, name:{}, ak:{}, cluster_mode:{}): '
            log_args = (
                sess_ctx.session_id,
                sess_ctx.session_type,
                sess_ctx.session_name,
                sess_ctx.access_key,
                sess_ctx.cluster_mode,
            )
            _log_fmt.set(log_fmt)
            _log_args.set(log_args)
            log.debug(log_fmt + 'try-scheduling', *log_args)
            session_agent_binding: Tuple[PendingSession, List[KernelAgentBinding]]

            async with kernel_db_conn.begin():
                predicates: Sequence[Tuple[str, Awaitable[PredicateResult]]] = [
                    (
                        'reserved_time',
                        check_reserved_batch_session(kernel_db_conn, sched_ctx, sess_ctx),
                    ),
                    ('concurrency', check_concurrency(kernel_db_conn, sched_ctx, sess_ctx)),
                    ('dependencies', check_dependencies(kernel_db_conn, sched_ctx, sess_ctx)),
                    (
                        'keypair_resource_limit',
                        check_keypair_resource_limit(kernel_db_conn, sched_ctx, sess_ctx),
                    ),
                    (
                        'user_group_resource_limit',
                        check_group_resource_limit(kernel_db_conn, sched_ctx, sess_ctx),
                    ),
                    (
                        'domain_resource_limit',
                        check_domain_resource_limit(kernel_db_conn, sched_ctx, sess_ctx),
                    ),
                    (
                        'scaling_group_resource_limit',
                        check_scaling_group(kernel_db_conn, sched_ctx, sess_ctx),
                    ),
                ]
                check_results: List[Tuple[str, Union[Exception, PredicateResult]]] = []
                for predicate_name, check_coro in predicates:
                    try:
                        check_results.append((predicate_name, await check_coro))
                    except Exception as e:
                        log.exception(log_fmt + 'predicate-error', *log_args)
                        check_results.append((predicate_name, e))
            has_failure = False
            # has_permanent_failure = False
            failed_predicates = []
            passed_predicates = []
            for predicate_name, result in check_results:
                if isinstance(result, Exception):
                    has_failure = True
                    failed_predicates.append({
                        'name': predicate_name,
                        'msg': repr(result),
                    })
                    continue
                if result.passed:
                    passed_predicates.append({
                        'name': predicate_name,
                    })
                else:
                    failed_predicates.append({
                        'name': predicate_name,
                        'msg': result.message or "",
                    })
                    has_failure = True
                    # if result.permanent:
                    #     has_permanent_failure = True
            if has_failure:
                log.debug(log_fmt + 'predicate-checks-failed (temporary)', *log_args)
                # TODO: handle has_permanent_failure as cancellation
                #  - An early implementation of it has caused DB query blocking due to
                #    the inclusion of the kernels.status field. :(
                #    Let's fix it.
                async with kernel_db_conn.begin():
                    await _invoke_failure_callbacks(
                        kernel_db_conn, sched_ctx, sess_ctx, check_results,
                    )
                    query = kernels.update().values({
                        'status_info': "predicate-checks-failed",
                        'status_data': sql_json_increment(
                            kernels.c.status_data,
                            ('scheduler', 'retries'),
                            parent_updates={
                                'last_try': datetime.now(tzutc()).isoformat(),
                                'failed_predicates': failed_predicates,
                                'passed_predicates': passed_predicates,
                            }
                        ),
                    }).where(kernels.c.id == sess_ctx.session_id)
                    await kernel_db_conn.execute(query)
                # Predicate failures are *NOT* permanent errors.
                # We need to retry the scheduling afterwards.
                continue
            else:
                async with kernel_db_conn.begin():
                    query = kernels.update().values({
                        'status_data': sql_json_merge(
                            kernels.c.status_data,
                            ('scheduler',),
                            {
                                'last_try': datetime.now(tzutc()).isoformat(),
                                'failed_predicates': failed_predicates,
                                'passed_predicates': passed_predicates,
                            }
                        ),
                    }).where(kernels.c.id == sess_ctx.session_id)
                    await kernel_db_conn.execute(query)

            if sess_ctx.cluster_mode == ClusterMode.SINGLE_NODE:
                session_agent_binding = await self._schedule_single_node_session(
                    sched_ctx,
                    scheduler,
                    agent_db_conn,
                    kernel_db_conn,
                    sgroup_name,
                    candidate_agents,
                    sess_ctx,
                    check_results,
                )
            elif sess_ctx.cluster_mode == ClusterMode.MULTI_NODE:
                session_agent_binding = await self._schedule_multi_node_session(
                    sched_ctx,
                    scheduler,
                    agent_db_conn,
                    kernel_db_conn,
                    sgroup_name,
                    candidate_agents,
                    sess_ctx,
                    check_results,
                )
            else:
                raise RuntimeError(
                    f"should not reach here; unknown cluster_mode: {sess_ctx.cluster_mode}"
                )
            args_list.append((
                log_args,
                sched_ctx,
                session_agent_binding,
                check_results,
            ))

        return args_list

    async def _schedule_single_node_session(
        self,
        sched_ctx: SchedulingContext,
        scheduler: AbstractScheduler,
        agent_db_conn: SAConnection,
        kernel_db_conn: SAConnection,
        sgroup_name: str,
        candidate_agents: Sequence[AgentContext],
        sess_ctx: PendingSession,
        check_results: List[Tuple[str, Union[Exception, PredicateResult]]],
    ) -> Tuple[PendingSession, List[KernelAgentBinding]]:
        # Assign agent resource per session.
        log_fmt = _log_fmt.get()
        log_args = _log_args.get()
        try:
            agent_id = scheduler.assign_agent_for_session(candidate_agents, sess_ctx)
            if agent_id is None:
                raise InstanceNotAvailable
            async with agent_db_conn.begin():
                agent_alloc_ctx = await _reserve_agent(
                    sched_ctx, agent_db_conn, sgroup_name, agent_id, sess_ctx.requested_slots,
                )
        except InstanceNotAvailable:
            log.debug(log_fmt + 'no-available-instances', *log_args)
            async with kernel_db_conn.begin():
                await _invoke_failure_callbacks(
                    kernel_db_conn, sched_ctx, sess_ctx, check_results,
                )
                query = kernels.update().values({
                    'status_info': "no-available-instances",
                    'status_data': sql_json_increment(
                        kernels.c.status_data,
                        ('scheduler', 'retries'),
                        parent_updates={
                            'last_try': datetime.now(tzutc()).isoformat(),
                        }
                    ),
                }).where(kernels.c.id == sess_ctx.session_id)
                await kernel_db_conn.execute(query)
            raise
        except Exception as e:
            log.exception(
                log_fmt + 'unexpected-error, during agent allocation',
                *log_args,
            )
            async with kernel_db_conn.begin():
                await _invoke_failure_callbacks(
                    kernel_db_conn, sched_ctx, sess_ctx, check_results,
                )
                query = kernels.update().values({
                    'status_info': "scheduler-error",
                    'status_data': convert_to_status_data(e),
                }).where(kernels.c.id == sess_ctx.session_id)
                await kernel_db_conn.execute(query)
            raise

        async with kernel_db_conn.begin():
            query = kernels.update().values({
                'agent': agent_alloc_ctx.agent_id,
                'agent_addr': agent_alloc_ctx.agent_addr,
                'scaling_group': sgroup_name,
                'status': KernelStatus.PREPARING,
                'status_info': 'scheduled',
                'status_data': {},
                'status_changed': datetime.now(tzutc()),
            }).where(kernels.c.session_id == sess_ctx.session_id)
            await kernel_db_conn.execute(query)
        return (
            sess_ctx,
            [
                KernelAgentBinding(kernel, agent_alloc_ctx)
                for kernel in sess_ctx.kernels
            ],
        )

    async def _schedule_multi_node_session(
        self,
        sched_ctx: SchedulingContext,
        scheduler: AbstractScheduler,
        agent_db_conn: SAConnection,
        kernel_db_conn: SAConnection,
        sgroup_name: str,
        candidate_agents: Sequence[AgentContext],
        sess_ctx: PendingSession,
        check_results: List[Tuple[str, Union[Exception, PredicateResult]]],
    ) -> Tuple[PendingSession, List[KernelAgentBinding]]:
        # Assign agent resource per kernel in the session.
        log_fmt = _log_fmt.get()
        log_args = _log_args.get()
        agent_query_extra_conds = None
        kernel_agent_bindings: List[KernelAgentBinding] = []
        async with agent_db_conn.begin(isolation_level="REPEATABLE READ"):
            # This outer transaction is rolled back when any exception occurs inside,
            # including scheduling failures of a kernel.
            # It ensures that occupied_slots are recovered when there are partial
            # scheduling failures.
            for kernel in sess_ctx.kernels:
                try:
                    agent_id = scheduler.assign_agent_for_kernel(candidate_agents, kernel)
                    if agent_id is None:
                        raise InstanceNotAvailable
                    async with agent_db_conn.begin_nested():
                        agent_alloc_ctx = await _reserve_agent(
                            sched_ctx, agent_db_conn,
                            sgroup_name, agent_id, kernel.requested_slots,
                            extra_conds=agent_query_extra_conds,
                        )
                        candidate_agents = await _list_agents_by_sgroup(agent_db_conn, sgroup_name)
                except InstanceNotAvailable:
                    log.debug(log_fmt + 'no-available-instances', *log_args)
                    async with kernel_db_conn.begin():
                        await _invoke_failure_callbacks(
                            kernel_db_conn, sched_ctx, sess_ctx, check_results,
                        )
                        query = kernels.update().values({
                            'status_info': "no-available-instances",
                            'status_data': sql_json_increment(
                                kernels.c.status_data,
                                ('scheduler', 'retries'),
                                parent_updates={
                                    'last_try': datetime.now(tzutc()).isoformat(),
                                }
                            ),
                        }).where(kernels.c.id == kernel.kernel_id)
                        await kernel_db_conn.execute(query)
                    raise
                except Exception as e:
                    log.exception(
                        log_fmt + 'unexpected-error, during agent allocation',
                        *log_args,
                    )
                    async with kernel_db_conn.begin():
                        await _invoke_failure_callbacks(
                            kernel_db_conn, sched_ctx, sess_ctx, check_results,
                        )
                        query = kernels.update().values({
                            'status_info': "scheduler-error",
                            'status_data': convert_to_status_data(e),
                        }).where(kernels.c.id == kernel.kernel_id)
                        await kernel_db_conn.execute(query)
                    raise
                else:
                    kernel_agent_bindings.append(KernelAgentBinding(kernel, agent_alloc_ctx))

        if len(kernel_agent_bindings) == len(sess_ctx.kernels):
            # Proceed to PREPARING only when all kernels are successfully scheduled.
            async with kernel_db_conn.begin():
                for binding in kernel_agent_bindings:
                    query = kernels.update().values({
                        'agent': binding.agent_alloc_ctx.agent_id,
                        'agent_addr': binding.agent_alloc_ctx.agent_addr,
                        'scaling_group': sgroup_name,
                        'status': KernelStatus.PREPARING,
                        'status_info': 'scheduled',
                        'status_data': {},
                        'status_changed': datetime.now(tzutc()),
                    }).where(kernels.c.id == binding.kernel.kernel_id)
                    await kernel_db_conn.execute(query)

        return (sess_ctx, kernel_agent_bindings)

    async def start_session(
        self,
        log_args,
        sched_ctx: SchedulingContext,
        session_agent_binding: Tuple[PendingSession, List[KernelAgentBinding]],
        check_results: List[Tuple[str, Union[Exception, PredicateResult]]],
    ) -> None:
        log_fmt = _log_fmt.get()
        log.debug(log_fmt + 'try-starting', *log_args)
        sess_ctx = session_agent_binding[0]
        await self.registry.event_producer.produce_event(
            SessionScheduledEvent(sess_ctx.session_id, sess_ctx.session_creation_id)
        )
        try:
            assert len(session_agent_binding[1]) > 0
            assert len(sess_ctx.kernels) == len(session_agent_binding[1])
            await self.registry.start_session(sched_ctx, session_agent_binding)
        except Exception as e:
            status_data = convert_to_status_data(e, self.local_config['debug']['enabled'])
            log.warning(log_fmt + 'failed-starting: {!r}', *log_args, status_data)
            try:
                async with self.db.begin() as db_conn:
                    query = (
                        sa.select([kernels.c.id, kernels.c.container_id])
                        .select_from(kernels)
                        .where(kernels.c.session_id == sess_ctx.session_id)
                    )
                    rows = (await db_conn.execute(query)).fetchall()
                    cid_map = {row['id']: row['container_id'] for row in rows}
                destroyed_kernels = [
                    {
                        "agent": binding.agent_alloc_ctx.agent_id,
                        "agent_addr": binding.agent_alloc_ctx.agent_addr,
                        "id": binding.kernel.kernel_id,
                        "container_id": cid_map[binding.kernel.kernel_id],
                    }
                    for binding in session_agent_binding[1]
                ]
                await self.registry.destroy_session_lowlevel(sess_ctx.session_id, destroyed_kernels)
            except Exception as destroy_err:
                log.error(log_fmt + 'failed-starting.cleanup', *log_args, exc_info=destroy_err)
            async with self.db.begin() as db_conn:
                for binding in session_agent_binding[1]:
                    await recalc_agent_resource_occupancy(db_conn, binding.agent_alloc_ctx.agent_id)
                await _invoke_failure_callbacks(db_conn, sched_ctx, sess_ctx, check_results)
                now = datetime.now(tzutc())
                query = kernels.update().values({
                    'status': KernelStatus.CANCELLED,
                    'status_changed': now,
                    'status_info': "failed-to-start",
                    'status_data': status_data,
                    'terminated_at': now,
                }).where(kernels.c.session_id == sess_ctx.session_id)
                await db_conn.execute(query)
            await self.registry.event_producer.produce_event(
                SessionCancelledEvent(
                    sess_ctx.session_id, sess_ctx.session_creation_id, "failed-to-start",
                )
            )
        else:
            log.info(log_fmt + 'started', *log_args)
            async with self.db.begin() as db_conn:
                await _invoke_success_callbacks(db_conn, sched_ctx, sess_ctx, check_results)


async def _list_pending_sessions(
    db_conn: SAConnection,
    sgroup_name: str,
) -> List[PendingSession]:
    query = (
        sa.select([
            kernels.c.id,
            kernels.c.session_creation_id,
            kernels.c.status,
            kernels.c.image,
            kernels.c.cluster_mode,
            kernels.c.cluster_size,
            kernels.c.cluster_role,
            kernels.c.cluster_idx,
            kernels.c.cluster_hostname,
            kernels.c.registry,
            kernels.c.session_id,
            kernels.c.session_type,
            kernels.c.session_name,
            kernels.c.access_key,
            kernels.c.domain_name,
            kernels.c.group_id,
            kernels.c.scaling_group,
            kernels.c.occupied_slots,
            kernels.c.resource_opts,
            kernels.c.environ,
            kernels.c.mounts,
            kernels.c.mount_map,
            kernels.c.bootstrap_script,
            kernels.c.startup_command,
            kernels.c.internal_data,
            kernels.c.preopen_ports,
            keypairs.c.resource_policy,
        ])
        .select_from(sa.join(
            kernels, keypairs,
            keypairs.c.access_key == kernels.c.access_key
        ))
        .where(
            (kernels.c.status == KernelStatus.PENDING) &
            (
                (kernels.c.scaling_group == sgroup_name) |
                (kernels.c.scaling_group.is_(None))
            )
        )
        .order_by(kernels.c.created_at)
    )
    items: Dict[SessionId, PendingSession] = {}
    rows = (await db_conn.execute(query)).fetchall()
    for row in rows:
        if row['cluster_role'] == "main":
            session = PendingSession(
                kernels=[],
                access_key=row['access_key'],
                session_id=row['session_id'],
                session_creation_id=row['session_creation_id'],
                session_type=row['session_type'],
                session_name=row['session_name'],
                cluster_mode=row['cluster_mode'],
                cluster_size=row['cluster_size'],
                domain_name=row['domain_name'],
                group_id=row['group_id'],
                scaling_group=row['scaling_group'],
                resource_policy=row['resource_policy'],
                resource_opts={},
                requested_slots=ResourceSlot(),
                internal_data=row['internal_data'],
                target_sgroup_names=[],
                environ={
                    k: v for k, v
                    in map(lambda s: s.split('=', maxsplit=1), row['environ'])
                },
                mounts=row['mounts'],
                mount_map=row['mount_map'],
                bootstrap_script=row['bootstrap_script'],
                startup_command=row['startup_command'],
                preopen_ports=row['preopen_ports'],
            )
            items[row['session_id']] = session
    for row in rows:
        session = items[row['session_id']]
        session.kernels.append(KernelInfo(
            kernel_id=row['id'],
            session_id=row['session_id'],
            access_key=row['access_key'],
            cluster_role=row['cluster_role'],
            cluster_idx=row['cluster_idx'],
            cluster_hostname=row['cluster_hostname'],
            image_ref=ImageRef(row['image'], [row['registry']]),
            bootstrap_script=row['bootstrap_script'],
            startup_command=row['startup_command'],
            resource_opts=row['resource_opts'],
            requested_slots=row['occupied_slots'],
        ))
        session.requested_slots += row['occupied_slots']  # type: ignore
        merge_resource(session.resource_opts, row['resource_opts'])  # type: ignore
    return list(items.values())


async def _list_existing_sessions(
    db_conn: SAConnection,
    sgroup: str,
) -> List[ExistingSession]:
    query = (
        sa.select([
            kernels.c.id,
            kernels.c.status,
            kernels.c.image,
            kernels.c.cluster_mode,
            kernels.c.cluster_size,
            kernels.c.cluster_role,
            kernels.c.cluster_idx,
            kernels.c.cluster_hostname,
            kernels.c.registry,
            kernels.c.session_id,
            kernels.c.session_type,
            kernels.c.session_name,
            kernels.c.access_key,
            kernels.c.domain_name,
            kernels.c.group_id,
            kernels.c.scaling_group,
            kernels.c.occupied_slots,
            kernels.c.resource_opts,
            kernels.c.environ,
            kernels.c.mounts,
            kernels.c.mount_map,
            kernels.c.startup_command,
            kernels.c.internal_data,
            keypairs.c.resource_policy,
        ])
        .select_from(sa.join(
            kernels, keypairs,
            keypairs.c.access_key == kernels.c.access_key
        ))
        .where(
            (kernels.c.status.in_(AGENT_RESOURCE_OCCUPYING_KERNEL_STATUSES)) &
            (kernels.c.scaling_group == sgroup)
        )
        .order_by(kernels.c.created_at)
    )
    items: Dict[str, ExistingSession] = {}
    rows = (await db_conn.execute(query)).fetchall()
    for row in rows:
        if row['cluster_role'] == "main":
            session = ExistingSession(
                kernels=[],
                access_key=row['access_key'],
                session_id=row['session_id'],
                session_type=row['session_type'],
                session_name=row['session_name'],
                cluster_mode=row['cluster_mode'],
                cluster_size=row['cluster_size'],
                domain_name=row['domain_name'],
                group_id=row['group_id'],
                scaling_group=row['scaling_group'],
                occupying_slots=ResourceSlot(),
            )
            items[row['session_id']] = session
    for row in rows:
        session_id = row['session_id']
        if session_id not in items:
            # In some cases, sub containers are still RUNNING even though main container is TERMINATED.
            # To circumvent this edge case, we skip if main container is not registered in `items`.
            continue
        session = items[session_id]
        session.kernels.append(KernelInfo(  # type: ignore
            kernel_id=row['id'],
            session_id=row['session_id'],
            access_key=row['access_key'],
            cluster_role=row['cluster_role'],
            cluster_idx=row['cluster_idx'],
            cluster_hostname=row['cluster_hostname'],
            image_ref=ImageRef(row['image'], [row['registry']]),
            bootstrap_script=None,
            startup_command=None,
            resource_opts=row['resource_opts'],
            requested_slots=row['occupied_slots'],
        ))
        session.occupying_slots += row['occupied_slots']  # type: ignore
    return list(items.values())


async def _list_agents_by_sgroup(
    db_conn: SAConnection,
    sgroup_name: str,
) -> Sequence[AgentContext]:
    query = (
        sa.select([
            agents.c.id,
            agents.c.addr,
            agents.c.scaling_group,
            agents.c.available_slots,
            agents.c.occupied_slots,
        ])
        .select_from(agents)
        .where(
            (agents.c.status == AgentStatus.ALIVE) &
            (agents.c.scaling_group == sgroup_name) &
            (agents.c.schedulable == true())
        )
    )
    items = []
    for row in (await db_conn.execute(query)):
        item = AgentContext(
            row['id'],
            row['addr'],
            row['scaling_group'],
            row['available_slots'],
            row['occupied_slots'],
        )
        items.append(item)
    return items


async def _reserve_agent(
    sched_ctx: SchedulingContext,
    db_conn: SAConnection,
    scaling_group: str,
    agent_id: AgentId,
    requested_slots: ResourceSlot,
    extra_conds: Any = None,
) -> AgentAllocationContext:
    query = (
        sa.select([agents.c.occupied_slots])
        .select_from(agents)
        .where(agents.c.id == agent_id))
    if extra_conds is not None:
        query = query.where(extra_conds)
    current_occupied_slots = await db_conn.scalar(query)
    if current_occupied_slots is None:
        raise RuntimeError(f"No agent matching condition: {extra_conds}")
    query = (sa.update(agents)
               .values({
                   'occupied_slots': current_occupied_slots + requested_slots
               })
               .where(agents.c.id == agent_id))
    await db_conn.execute(query)

    # Get the agent address for later RPC calls
    query = (sa.select([agents.c.addr])
               .where(agents.c.id == agent_id))
    agent_addr = await db_conn.scalar(query)
    assert agent_addr is not None

    return AgentAllocationContext(agent_id, agent_addr, scaling_group)


async def _unreserve_agent_slots(
    db_conn: SAConnection,
    session_agent_binding: Tuple[PendingSession, List[KernelAgentBinding]],
) -> None:
    # Un-reserve agent slots, using the db transaction of the current invocation context.
    keyfunc = lambda item: item.agent_alloc_ctx.agent_id
    for agent_id, kernel_agent_bindings in itertools.groupby(
        sorted(session_agent_binding[1], key=keyfunc), key=keyfunc
    ):
        per_agent_requested_slots = sum(
            (binding.kernel.requested_slots for binding in kernel_agent_bindings),
            start=ResourceSlot(),
        )
        query = (
            sa.select([agents.c.occupied_slots])
            .select_from(agents)
            .where(agents.c.id == agent_id)
            .with_for_update()
        )
        current_occupied_slots = await db_conn.scalar(query)
        query = (
            sa.update(agents)
            .values({
                'occupied_slots': current_occupied_slots - per_agent_requested_slots
            })
            .where(agents.c.id == agent_id))
        await db_conn.execute(query)


async def _invoke_success_callbacks(
    db_conn: SAConnection,
    sched_ctx: SchedulingContext,
    sess_ctx: PendingSession,
    results: List[Tuple[str, Union[Exception, PredicateResult]]],
    *,
    use_new_txn: bool = False,
) -> None:
    """
    Give predicates chances to finalize/add DB changes.
    """
    callbacks: List[Awaitable[None]] = []
    for predicate_name, result in results:
        if isinstance(result, Exception):
            # This won't happen but this code is required to pass static check.
            continue
        if result.success_cb is not None:
            callbacks.append(result.success_cb(db_conn, sched_ctx, sess_ctx))
    for cb in reversed(callbacks):
        await cb


async def _invoke_failure_callbacks(
    db_conn: SAConnection,
    sched_ctx: SchedulingContext,
    sess_ctx: PendingSession,
    results: List[Tuple[str, Union[Exception, PredicateResult]]],
    *,
    use_new_txn: bool = False,
) -> None:
    """
    Rollback any changes performed by predicates.

    NOTE: We don't use the DB-level transaction rollback because we need to
    store the "ERROR" status to corresponding rows in the kernels table.
    """
    callbacks: List[Awaitable[None]] = []
    for predicate_name, result in results:
        if isinstance(result, Exception):
            # This won't happen but this code is required to pass static check.
            continue
        if result.failure_cb:
            callbacks.append(result.failure_cb(db_conn, sched_ctx, sess_ctx))
    for cb in reversed(callbacks):
        await cb



"""add-status_data

Revision ID: 250e8656cf45
Revises: 57e717103287
Create Date: 2020-12-23 14:19:08.801283

"""
from alembic import op
import sqlalchemy as sa
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = '250e8656cf45'
down_revision = '57e717103287'
branch_labels = None
depends_on = None


def upgrade():
    op.add_column('kernels', sa.Column('status_data', postgresql.JSONB(astext_type=sa.Text()), nullable=True))


def downgrade():
    op.drop_column('kernels', 'status_data')



"""vfolder_invitation_state_to_enum_type

Revision ID: 6f5fe19894b7
Revises: 48ab2dfefba9
Create Date: 2019-09-28 21:05:55.409422

"""
from alembic import op
import sqlalchemy as sa
from ai.backend.manager.models import VFolderInvitationState
from sqlalchemy.dialects import postgresql


# revision identifiers, used by Alembic.
revision = '6f5fe19894b7'
down_revision = '48ab2dfefba9'
branch_labels = None
depends_on = None

vfinvs_choices = list(map(lambda v: v.value, VFolderInvitationState))
vfolderinvitationstate = postgresql.ENUM(
    *vfinvs_choices,
    name='vfolderinvitationstate',
)


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    pass
    # ### end Alembic commands ###
    vfolderinvitationstate.create(op.get_bind())
    op.alter_column('vfolder_invitations', column_name='state',
                    type_=sa.Enum(*vfinvs_choices, name='vfolderinvitationstate'),
                    postgresql_using='state::vfolderinvitationstate')


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    pass
    # ### end Alembic commands ###
    vfolderinvitationstate.create(op.get_bind())
    op.alter_column('vfolder_invitations', column_name='state',
                    type_=sa.String(length=10),
                    postgresql_using='state::text::vfolderinvitationstate')
    vfolderinvitationstate.drop(op.get_bind())



from __future__ import annotations

import logging
from typing import (
    Any,
    Dict,
    Sequence,
    TYPE_CHECKING,
)

import graphene
from graphene.types.datetime import DateTime as GQLDateTime
import sqlalchemy as sa
from sqlalchemy.engine.row import Row
from sqlalchemy.dialects import postgresql as pgsql

from ai.backend.common.logging import BraceStyleAdapter
from ai.backend.common.types import DefaultForUnspecified, ResourceSlot
from .base import (
    metadata, BigInt, EnumType, ResourceSlotColumn,
    simple_db_mutate,
    simple_db_mutate_returning_item,
    set_if_set,
    batch_result,
)
from .keypair import keypairs
from .user import UserRole

if TYPE_CHECKING:
    from .gql import GraphQueryContext

log = BraceStyleAdapter(logging.getLogger('ai.backend.manager.models'))

__all__: Sequence[str] = (
    'keypair_resource_policies',
    'KeyPairResourcePolicy',
    'DefaultForUnspecified',
    'CreateKeyPairResourcePolicy',
    'ModifyKeyPairResourcePolicy',
    'DeleteKeyPairResourcePolicy',
)


keypair_resource_policies = sa.Table(
    'keypair_resource_policies', metadata,
    sa.Column('name', sa.String(length=256), primary_key=True),
    sa.Column('created_at', sa.DateTime(timezone=True),
              server_default=sa.func.now()),
    sa.Column('default_for_unspecified',
              EnumType(DefaultForUnspecified),
              default=DefaultForUnspecified.LIMITED,
              nullable=False),
    sa.Column('total_resource_slots', ResourceSlotColumn(), nullable=False),
    sa.Column('max_concurrent_sessions', sa.Integer(), nullable=False),
    sa.Column('max_containers_per_session', sa.Integer(), nullable=False),
    sa.Column('max_vfolder_count', sa.Integer(), nullable=False),
    sa.Column('max_vfolder_size', sa.BigInteger(), nullable=False),
    sa.Column('idle_timeout', sa.BigInteger(), nullable=False),
    sa.Column('allowed_vfolder_hosts', pgsql.ARRAY(sa.String), nullable=False),
    # TODO: implement with a many-to-many association table
    # sa.Column('allowed_scaling_groups', sa.Array(sa.String), nullable=False),
)


class KeyPairResourcePolicy(graphene.ObjectType):
    name = graphene.String()
    created_at = GQLDateTime()
    default_for_unspecified = graphene.String()
    total_resource_slots = graphene.JSONString()
    max_concurrent_sessions = graphene.Int()
    max_containers_per_session = graphene.Int()
    idle_timeout = BigInt()
    max_vfolder_count = graphene.Int()
    max_vfolder_size = BigInt()
    allowed_vfolder_hosts = graphene.List(lambda: graphene.String)

    @classmethod
    def from_row(
        cls,
        ctx: GraphQueryContext,
        row: Row | None,
    ) -> KeyPairResourcePolicy | None:
        if row is None:
            return None
        return cls(
            name=row['name'],
            created_at=row['created_at'],
            default_for_unspecified=row['default_for_unspecified'].name,
            total_resource_slots=row['total_resource_slots'].to_json(),
            max_concurrent_sessions=row['max_concurrent_sessions'],
            max_containers_per_session=row['max_containers_per_session'],
            idle_timeout=row['idle_timeout'],
            max_vfolder_count=row['max_vfolder_count'],
            max_vfolder_size=row['max_vfolder_size'],
            allowed_vfolder_hosts=row['allowed_vfolder_hosts'],
        )

    @classmethod
    async def load_all(cls, ctx: GraphQueryContext) -> Sequence[KeyPairResourcePolicy]:
        async with ctx.db.begin() as conn:
            query = (sa.select([keypair_resource_policies])
                       .select_from(keypair_resource_policies))
            return [
                obj async for r in (await conn.stream(query))
                if (obj := cls.from_row(ctx, r)) is not None
            ]

    @classmethod
    async def load_all_user(
        cls,
        ctx: GraphQueryContext,
        access_key: str,
    ) -> Sequence[KeyPairResourcePolicy]:
        async with ctx.db.begin() as conn:
            query = (
                sa.select([keypairs.c.user_id])
                .select_from(keypairs)
                .where(keypairs.c.access_key == access_key)
            )
            result = await conn.execute(query)
            row = result.first()
            user_id = row['user_id']
            j = sa.join(
                keypairs, keypair_resource_policies,
                keypairs.c.resource_policy == keypair_resource_policies.c.name
            )
            query = (
                sa.select([keypair_resource_policies])
                .select_from(j)
                .where((keypairs.c.user_id == user_id))
            )
            return [
                obj async for r in (await conn.stream(query))
                if (obj := cls.from_row(ctx, r)) is not None
            ]

    @classmethod
    async def batch_load_by_name(
        cls,
        ctx: GraphQueryContext,
        names: Sequence[str],
    ) -> Sequence[KeyPairResourcePolicy | None]:
        async with ctx.db.begin() as conn:
            query = (
                sa.select([keypair_resource_policies])
                .select_from(keypair_resource_policies)
                .where(keypair_resource_policies.c.name.in_(names))
                .order_by(keypair_resource_policies.c.name)
            )
            return await batch_result(
                ctx, conn, query, cls,
                names, lambda row: row['name'],
            )

    @classmethod
    async def batch_load_by_name_user(
        cls,
        ctx: GraphQueryContext,
        names: Sequence[str],
    ) -> Sequence[KeyPairResourcePolicy | None]:
        async with ctx.db.begin() as conn:
            access_key = ctx.access_key
            j = sa.join(
                keypairs, keypair_resource_policies,
                keypairs.c.resource_policy == keypair_resource_policies.c.name
            )
            query = (
                sa.select([keypair_resource_policies])
                .select_from(j)
                .where(
                    (keypair_resource_policies.c.name.in_(names)) &
                    (keypairs.c.access_key == access_key)
                )
                .order_by(keypair_resource_policies.c.name))
            return await batch_result(
                ctx, conn, query, cls,
                names, lambda row: row['name'],
            )

    @classmethod
    async def batch_load_by_ak(
        cls,
        ctx: GraphQueryContext,
        access_keys: Sequence[str],
    ) -> Sequence[KeyPairResourcePolicy]:
        async with ctx.db.begin() as conn:
            j = sa.join(
                keypairs, keypair_resource_policies,
                keypairs.c.resource_policy == keypair_resource_policies.c.name
            )
            query = (sa.select([keypair_resource_policies])
                       .select_from(j)
                       .where((keypairs.c.access_key.in_(access_keys)))
                       .order_by(keypair_resource_policies.c.name))
            return [
                obj async for r in (await conn.stream(query))
                if (obj := cls.from_row(ctx, r)) is not None
            ]


class CreateKeyPairResourcePolicyInput(graphene.InputObjectType):
    default_for_unspecified = graphene.String(required=True)
    total_resource_slots = graphene.JSONString(required=True)
    max_concurrent_sessions = graphene.Int(required=True)
    max_containers_per_session = graphene.Int(required=True)
    idle_timeout = BigInt(required=True)
    max_vfolder_count = graphene.Int(required=True)
    max_vfolder_size = BigInt(required=True)
    allowed_vfolder_hosts = graphene.List(lambda: graphene.String)


class ModifyKeyPairResourcePolicyInput(graphene.InputObjectType):
    default_for_unspecified = graphene.String(required=False)
    total_resource_slots = graphene.JSONString(required=False)
    max_concurrent_sessions = graphene.Int(required=False)
    max_containers_per_session = graphene.Int(required=False)
    idle_timeout = BigInt(required=False)
    max_vfolder_count = graphene.Int(required=False)
    max_vfolder_size = BigInt(required=False)
    allowed_vfolder_hosts = graphene.List(lambda: graphene.String, required=False)


class CreateKeyPairResourcePolicy(graphene.Mutation):

    allowed_roles = (UserRole.SUPERADMIN,)

    class Arguments:
        name = graphene.String(required=True)
        props = CreateKeyPairResourcePolicyInput(required=True)

    ok = graphene.Boolean()
    msg = graphene.String()
    resource_policy = graphene.Field(lambda: KeyPairResourcePolicy, required=False)

    @classmethod
    async def mutate(
        cls,
        root,
        info: graphene.ResolveInfo,
        name: str,
        props: CreateKeyPairResourcePolicyInput,
    ) -> CreateKeyPairResourcePolicy:
        data = {
            'name': name,
            'default_for_unspecified':
                DefaultForUnspecified[props.default_for_unspecified],
            'total_resource_slots': ResourceSlot.from_user_input(
                props.total_resource_slots, None),
            'max_concurrent_sessions': props.max_concurrent_sessions,
            'max_containers_per_session': props.max_containers_per_session,
            'idle_timeout': props.idle_timeout,
            'max_vfolder_count': props.max_vfolder_count,
            'max_vfolder_size': props.max_vfolder_size,
            'allowed_vfolder_hosts': props.allowed_vfolder_hosts,
        }
        insert_query = (keypair_resource_policies.insert().values(data))
        item_query = (
            keypair_resource_policies.select()
            .where(keypair_resource_policies.c.name == name)
        )
        return await simple_db_mutate_returning_item(
            cls, info.context, insert_query,
            item_query=item_query, item_cls=KeyPairResourcePolicy)


class ModifyKeyPairResourcePolicy(graphene.Mutation):

    allowed_roles = (UserRole.SUPERADMIN,)

    class Arguments:
        name = graphene.String(required=True)
        props = ModifyKeyPairResourcePolicyInput(required=True)

    ok = graphene.Boolean()
    msg = graphene.String()

    @classmethod
    async def mutate(
        cls,
        root,
        info: graphene.ResolveInfo,
        name: str,
        props: ModifyKeyPairResourcePolicyInput,
    ) -> ModifyKeyPairResourcePolicy:
        data: Dict[str, Any] = {}
        set_if_set(props, data, 'default_for_unspecified',
                   clean_func=lambda v: DefaultForUnspecified[v])
        set_if_set(props, data, 'total_resource_slots',
                   clean_func=lambda v: ResourceSlot.from_user_input(v, None))
        set_if_set(props, data, 'max_concurrent_sessions')
        set_if_set(props, data, 'max_containers_per_session')
        set_if_set(props, data, 'idle_timeout')
        set_if_set(props, data, 'max_vfolder_count')
        set_if_set(props, data, 'max_vfolder_size')
        set_if_set(props, data, 'allowed_vfolder_hosts')
        update_query = (
            keypair_resource_policies.update()
            .values(data)
            .where(keypair_resource_policies.c.name == name))
        return await simple_db_mutate(cls, info.context, update_query)


class DeleteKeyPairResourcePolicy(graphene.Mutation):

    allowed_roles = (UserRole.SUPERADMIN,)

    class Arguments:
        name = graphene.String(required=True)

    ok = graphene.Boolean()
    msg = graphene.String()

    @classmethod
    async def mutate(
        cls,
        root,
        info: graphene.ResolveInfo,
        name: str,
    ) -> DeleteKeyPairResourcePolicy:
        delete_query = (
            keypair_resource_policies.delete()
            .where(keypair_resource_policies.c.name == name)
        )
        return await simple_db_mutate(cls, info.context, delete_query)



from __future__ import annotations

from decimal import Decimal
import logging
import time
from typing import (
    Iterable,
    Final,
    Tuple,
)

from aiohttp import web
from aiotools import apartial
from aioredis import Redis
import attr

from ai.backend.common import redis
from ai.backend.common.logging import BraceStyleAdapter

from ..defs import REDIS_RLIM_DB
from .context import RootContext
from .exceptions import RateLimitExceeded
from .types import CORSOptions, WebRequestHandler, WebMiddleware

log = BraceStyleAdapter(logging.getLogger(__name__))

_time_prec: Final = Decimal('1e-3')  # msec
_rlim_window: Final = 60 * 15

# We implement rate limiting using a rolling counter, which prevents
# last-minute and first-minute bursts between the intervals.

_rlim_script = '''
local access_key = KEYS[1]
local now = tonumber(ARGV[1])
local window = tonumber(ARGV[2])
local request_id = tonumber(redis.call('INCR', '__request_id'))
if request_id >= 1e12 then
    redis.call('SET', '__request_id', 1)
end
if redis.call('EXISTS', access_key) == 1 then
    redis.call('ZREMRANGEBYSCORE', access_key, 0, now - window)
end
redis.call('ZADD', access_key, now, tostring(request_id))
redis.call('EXPIRE', access_key, window)
return redis.call('ZCARD', access_key)
'''


@web.middleware
async def rlim_middleware(
    app: web.Application,
    request: web.Request,
    handler: WebRequestHandler,
) -> web.StreamResponse:
    # This is a global middleware: request.app is the root app.
    app_ctx: PrivateContext = app['ratelimit.context']
    now = Decimal(time.time()).quantize(_time_prec)
    rr = app_ctx.redis_rlim
    if request['is_authorized']:
        rate_limit = request['keypair']['rate_limit']
        access_key = request['keypair']['access_key']
        ret = await redis.execute_script(
            rr, 'ratelimit', _rlim_script,
            [access_key],
            [str(now), str(_rlim_window)],
        )
        if ret is None:
            remaining = rate_limit
        else:
            rolling_count = int(ret)
            if rolling_count > rate_limit:
                raise RateLimitExceeded
            remaining = rate_limit - rolling_count
        response = await handler(request)
        response.headers['X-RateLimit-Limit'] = str(rate_limit)
        response.headers['X-RateLimit-Remaining'] = str(remaining)
        response.headers['X-RateLimit-Window'] = str(_rlim_window)
        return response
    else:
        # No checks for rate limiting for non-authorized queries.
        response = await handler(request)
        response.headers['X-RateLimit-Limit'] = '1000'
        response.headers['X-RateLimit-Remaining'] = '1000'
        response.headers['X-RateLimit-Window'] = str(_rlim_window)
        return response


@attr.s(slots=True, auto_attribs=True, init=False)
class PrivateContext:
    redis_rlim: Redis
    redis_rlim_script: str


async def init(app: web.Application) -> None:
    root_ctx: RootContext = app['_root.context']
    app_ctx: PrivateContext = app['ratelimit.context']
    rr = await redis.connect_with_retries(
        str(root_ctx.shared_config.get_redis_url(db=REDIS_RLIM_DB)),
        timeout=3.0,
        encoding='utf8',
    )
    app_ctx.redis_rlim = rr
    app_ctx.redis_rlim_script = await rr.script_load(_rlim_script)


async def shutdown(app: web.Application) -> None:
    app_ctx: PrivateContext = app['ratelimit.context']
    try:
        await app_ctx.redis_rlim.flushdb()
    except (ConnectionResetError, ConnectionRefusedError):
        pass
    app_ctx.redis_rlim.close()
    await app_ctx.redis_rlim.wait_closed()


def create_app(default_cors_options: CORSOptions) -> Tuple[web.Application, Iterable[WebMiddleware]]:
    app = web.Application()
    app['api_versions'] = (1, 2, 3, 4)
    app['ratelimit.context'] = PrivateContext()
    app.on_startup.append(init)
    app.on_shutdown.append(shutdown)
    # middleware must be wrapped by web.middleware at the outermost level.
    return app, [web.middleware(apartial(rlim_middleware, app))]



"""Update for multi-container sessions.

Revision ID: d5cc54fd36b5
Revises: 0d553d59f369
Create Date: 2020-01-06 13:56:50.885635

"""
from alembic import op
import sqlalchemy as sa

from ai.backend.manager.models.base import GUID

# revision identifiers, used by Alembic.
revision = 'd5cc54fd36b5'
down_revision = '0d553d59f369'
branch_labels = None
depends_on = None


def upgrade():
    # In this mgiration, we finally clear up the column namings:
    #   sess_id -> session_name
    #     => client-provided alias
    #   (new) -> session_id
    #     => for single-container sessions, it may be derived from the kernel id.
    #   sess_type -> session_type
    #
    op.drop_index('ix_kernels_sess_id', table_name='kernels')
    op.drop_index('ix_kernels_sess_type', table_name='kernels')

    conn = op.get_bind()
    op.add_column(
        'kernels',
        sa.Column('idx', sa.Integer, nullable=True, default=None))
    op.add_column(
        'kernels',
        sa.Column('cluster_mode', sa.String(16), nullable=False,
                  default='single-node', server_default='single-node'))

    # Set idx to 1 (previous sessions are all composed of one kernel)
    query = "UPDATE kernels SET idx = 1;"
    conn.execute(query)

    # Convert "master" to "main"
    # NOTE: "main" is defined from ai.backend.manager.defs.DEFAULT_ROLE
    op.alter_column('kernels', 'role', server_default='main')
    query = "UPDATE kernels SET role = 'main' WHERE role = 'master'"
    conn.execute(query)

    # First a session_id column as nullable and fill it up before setting it non-nullable.
    op.add_column('kernels', sa.Column('session_id', GUID, nullable=True))
    query = "UPDATE kernels SET session_id = kernels.id WHERE role = 'main'"
    conn.execute(query)
    # If we upgrade from a database downgraded in the past with sub-kernel records,
    # we loose the information of kernel_id -> session_id mapping.
    # Try to restore it by getting the session ID of a main-kernel record which is created
    # at a similar time range.  This will raise an error if there are two or more such records,
    # and it is based on an assumption that development setups with manual tests would not make such
    # overlaps.
    query = """
    UPDATE kernels t SET session_id = (
        SELECT session_id
        FROM kernels s
        WHERE
            s.role = 'main'
            AND (
                s.created_at BETWEEN
                t.created_at - (interval '0.5s')
                AND t.created_at + (interval '3s')
            )
    )
    WHERE t.role <> 'main'
    """
    conn.execute(query)
    op.alter_column('kernels', 'session_id', nullable=False)

    op.alter_column('kernels', 'sess_id', new_column_name='session_name')
    op.alter_column('kernels', 'sess_type', new_column_name='session_type')

    op.create_index(op.f('ix_kernels_session_id'), 'kernels', ['session_id'], unique=False)
    op.create_index(op.f('ix_kernels_session_name'), 'kernels', ['session_name'], unique=False)
    op.create_index(op.f('ix_kernels_session_type'), 'kernels', ['session_type'], unique=False)


def downgrade():
    op.drop_index(op.f('ix_kernels_session_type'), table_name='kernels')
    op.drop_index(op.f('ix_kernels_session_name'), table_name='kernels')
    op.drop_index(op.f('ix_kernels_session_id'), table_name='kernels')

    op.alter_column('kernels', 'session_type', new_column_name='sess_type')
    op.alter_column('kernels', 'session_name', new_column_name='sess_id')
    op.drop_column('kernels', 'session_id')

    # Convert "main" to "master" for backward compatibility
    op.alter_column('kernels', 'role', server_default='master')
    conn = op.get_bind()
    query = "UPDATE kernels SET role = 'master' WHERE role = 'main'"
    conn.execute(query)

    op.drop_column('kernels', 'cluster_mode')
    op.drop_column('kernels', 'idx')

    op.create_index('ix_kernels_sess_type', 'kernels', ['sess_type'], unique=False)
    op.create_index('ix_kernels_sess_id', 'kernels', ['sess_id'], unique=False)



from __future__ import annotations

import asyncio
import base64
import secrets
from typing import (
    Any,
    Dict,
    Optional,
    Sequence,
    List, TYPE_CHECKING,
    Tuple,
    TypedDict,
)
import uuid

from cryptography.hazmat.primitives import serialization as crypto_serialization
from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.hazmat.backends import default_backend as crypto_default_backend
import graphene
from graphene.types.datetime import DateTime as GQLDateTime
import sqlalchemy as sa
from sqlalchemy.ext.asyncio import AsyncConnection as SAConnection
from sqlalchemy.engine.row import Row
from sqlalchemy.sql.expression import false

from ai.backend.common import msgpack
from ai.backend.common.types import (
    AccessKey,
    SecretKey,
)

if TYPE_CHECKING:
    from .gql import GraphQueryContext
    from .vfolder import VirtualFolder

from .base import (
    ForeignKeyIDColumn,
    Item,
    PaginatedList,
    metadata,
    batch_result,
    batch_multiresult,
    set_if_set,
    simple_db_mutate,
)
from .user import ModifyUserInput, UserRole
from ..defs import RESERVED_DOTFILES

__all__: Sequence[str] = (
    'keypairs',
    'KeyPair', 'KeyPairList',
    'UserInfo',
    'KeyPairInput',
    'CreateKeyPair', 'ModifyKeyPair', 'DeleteKeyPair',
    'Dotfile', 'MAXIMUM_DOTFILE_SIZE',
    'query_owned_dotfiles',
    'query_bootstrap_script',
    'verify_dotfile_name'
)


MAXIMUM_DOTFILE_SIZE = 64 * 1024  # 61 KiB

keypairs = sa.Table(
    'keypairs', metadata,
    sa.Column('user_id', sa.String(length=256), index=True),
    sa.Column('access_key', sa.String(length=20), primary_key=True),
    sa.Column('secret_key', sa.String(length=40)),
    sa.Column('is_active', sa.Boolean, index=True),
    sa.Column('is_admin', sa.Boolean, index=True,
              default=False, server_default=false()),
    sa.Column('created_at', sa.DateTime(timezone=True),
              server_default=sa.func.now()),
    sa.Column('modified_at', sa.DateTime(timezone=True),
              server_default=sa.func.now(), onupdate=sa.func.current_timestamp()),
    sa.Column('last_used', sa.DateTime(timezone=True), nullable=True),
    sa.Column('concurrency_used', sa.Integer),
    sa.Column('rate_limit', sa.Integer),
    sa.Column('num_queries', sa.Integer, server_default='0'),

    # SSH Keypairs.
    sa.Column('ssh_public_key', sa.String(length=750), nullable=True),
    sa.Column('ssh_private_key', sa.String(length=2000), nullable=True),

    ForeignKeyIDColumn('user', 'users.uuid', nullable=False),
    sa.Column('resource_policy', sa.String(length=256),
              sa.ForeignKey('keypair_resource_policies.name'),
              nullable=False),
    # dotfiles column, \x90 means empty list in msgpack
    sa.Column('dotfiles', sa.LargeBinary(length=MAXIMUM_DOTFILE_SIZE), nullable=False, default=b'\x90'),
    sa.Column('bootstrap_script', sa.String(length=MAXIMUM_DOTFILE_SIZE), nullable=False, default=''),
)


class UserInfo(graphene.ObjectType):
    email = graphene.String()
    full_name = graphene.String()

    @classmethod
    def from_row(
        cls,
        ctx: GraphQueryContext,
        row: Row,
    ) -> Optional[UserInfo]:
        if row is None:
            return None
        return cls(email=row['email'], full_name=row['full_name'])

    @classmethod
    async def batch_load_by_uuid(
        cls,
        ctx: GraphQueryContext,
        user_uuids: Sequence[uuid.UUID],
    ) -> Sequence[Optional[UserInfo]]:
        async with ctx.db.begin() as conn:
            from .user import users
            query = (
                sa.select([users.c.uuid, users.c.email, users.c.full_name])
                .select_from(users)
                .where(users.c.uuid.in_(user_uuids))
            )
            return await batch_result(
                ctx, conn, query, cls,
                user_uuids, lambda row: row['uuid'],
            )


class KeyPair(graphene.ObjectType):

    class Meta:
        interfaces = (Item, )

    user_id = graphene.String()
    access_key = graphene.String()
    secret_key = graphene.String()
    is_active = graphene.Boolean()
    is_admin = graphene.Boolean()
    resource_policy = graphene.String()
    created_at = GQLDateTime()
    last_used = GQLDateTime()
    concurrency_used = graphene.Int()
    rate_limit = graphene.Int()
    num_queries = graphene.Int()
    user = graphene.UUID()

    ssh_public_key = graphene.String()

    vfolders = graphene.List('ai.backend.manager.models.VirtualFolder')
    compute_sessions = graphene.List(
        'ai.backend.manager.models.ComputeSession',
        status=graphene.String(),
    )

    user_info = graphene.Field(lambda: UserInfo)

    # Deprecated
    concurrency_limit = graphene.Int(
        deprecation_reason='Moved to KeyPairResourcePolicy object as '
                           'max_concurrent_sessions field.')

    async def resolve_user_info(
        self,
        info: graphene.ResolveInfo,
    ) -> UserInfo:
        ctx: GraphQueryContext = info.context
        loader = ctx.dataloader_manager.get_loader(ctx, 'UserInfo.by_uuid')
        return await loader.load(self.user)

    @classmethod
    def from_row(
        cls,
        ctx: GraphQueryContext,
        row: Row,
    ) -> KeyPair:
        return cls(
            id=row['access_key'],
            user_id=row['user_id'],
            access_key=row['access_key'],
            secret_key=row['secret_key'],
            is_active=row['is_active'],
            is_admin=row['is_admin'],
            resource_policy=row['resource_policy'],
            created_at=row['created_at'],
            last_used=row['last_used'],
            concurrency_limit=0,  # moved to resource policy
            concurrency_used=row['concurrency_used'],
            rate_limit=row['rate_limit'],
            num_queries=row['num_queries'],
            user=row['user'],
            ssh_public_key=row['ssh_public_key'],
        )

    async def resolve_vfolders(self, info: graphene.ResolveInfo) -> Sequence[VirtualFolder]:
        ctx: GraphQueryContext = info.context
        loader = ctx.dataloader_manager.get_loader(ctx, 'VirtualFolder')
        return await loader.load(self.access_key)

    async def resolve_compute_sessions(self, info: graphene.ResolveInfo, raw_status: str = None):
        ctx: GraphQueryContext = info.context
        from . import KernelStatus  # noqa: avoid circular imports
        if raw_status is not None:
            status = KernelStatus[raw_status]
        loader = ctx.dataloader_manager.get_loader(ctx, 'ComputeSession', status=status)
        return await loader.load(self.access_key)

    @classmethod
    async def load_all(
        cls,
        graph_ctx: GraphQueryContext,
        *,
        domain_name: str = None,
        is_active: bool = None,
        limit: int = None,
    ) -> Sequence[KeyPair]:
        from .user import users
        j = sa.join(
            keypairs, users,
            keypairs.c.user == users.c.uuid,
        )
        query = (
            sa.select([keypairs])
            .select_from(j)
        )
        if domain_name is not None:
            query = query.where(users.c.domain_name == domain_name)
        if is_active is not None:
            query = query.where(keypairs.c.is_active == is_active)
        if limit is not None:
            query = query.limit(limit)
        return [
            obj async for row in (await graph_ctx.db_conn.stream(query))
            if (obj := cls.from_row(graph_ctx, row)) is not None
        ]

    @staticmethod
    async def load_count(
        graph_ctx: GraphQueryContext,
        *,
        domain_name: str = None,
        email: str = None,
        is_active: bool = None,
    ) -> int:
        from .user import users
        j = sa.join(keypairs, users, keypairs.c.user == users.c.uuid)
        query = (
            sa.select([sa.func.count(keypairs.c.access_key)])
            .select_from(j)
        )
        if domain_name is not None:
            query = query.where(users.c.domain_name == domain_name)
        if email is not None:
            query = query.where(keypairs.c.user_id == email)
        if is_active is not None:
            query = query.where(keypairs.c.is_active == is_active)
        result = await graph_ctx.db_conn.execute(query)
        return result.scalar()

    @classmethod
    async def load_slice(
        cls,
        graph_ctx: GraphQueryContext,
        limit: int,
        offset: int,
        *,
        domain_name: str = None,
        email: str = None,
        is_active: bool = None,
        order_key: str = None,
        order_asc: bool = True,
    ) -> Sequence[KeyPair]:
        from .user import users
        if order_key is None:
            _ordering = sa.desc(keypairs.c.created_at)
        else:
            _order_func = sa.asc if order_asc else sa.desc
            _ordering = _order_func(getattr(keypairs.c, order_key))
        j = sa.join(keypairs, users, keypairs.c.user == users.c.uuid)
        query = (
            sa.select([keypairs])
            .select_from(j)
            .order_by(_ordering)
            .limit(limit)
            .offset(offset)
        )
        if domain_name is not None:
            query = query.where(users.c.domain_name == domain_name)
        if email is not None:
            query = query.where(keypairs.c.user_id == email)
        if is_active is not None:
            query = query.where(keypairs.c.is_active == is_active)
        return [
            obj async for row in (await graph_ctx.db_conn.stream(query))
            if (obj := cls.from_row(graph_ctx, row)) is not None
        ]

    @classmethod
    async def batch_load_by_email(
        cls,
        graph_ctx: GraphQueryContext,
        user_ids: Sequence[uuid.UUID],
        *,
        domain_name: str = None,
        is_active: bool = None,
    ) -> Sequence[Sequence[Optional[KeyPair]]]:
        from .user import users
        j = sa.join(
            keypairs, users,
            keypairs.c.user == users.c.uuid,
        )
        query = (
            sa.select([keypairs])
            .select_from(j)
            .where(keypairs.c.user_id.in_(user_ids))
        )
        if domain_name is not None:
            query = query.where(users.c.domain_name == domain_name)
        if is_active is not None:
            query = query.where(keypairs.c.is_active == is_active)
        return await batch_multiresult(
            graph_ctx, graph_ctx.db_conn, query, cls,
            user_ids, lambda row: row['user_id'],
        )

    @classmethod
    async def batch_load_by_ak(
        cls,
        graph_ctx: GraphQueryContext,
        access_keys: Sequence[AccessKey],
        *,
        domain_name: str = None,
    ) -> Sequence[Optional[KeyPair]]:
        from .user import users
        j = sa.join(
            keypairs, users,
            keypairs.c.user == users.c.uuid,
        )
        query = (
            sa.select([keypairs])
            .select_from(j)
            .where(keypairs.c.access_key.in_(access_keys))
        )
        if domain_name is not None:
            query = query.where(users.c.domain_name == domain_name)
        return await batch_result(
            graph_ctx, graph_ctx.db_conn, query, cls,
            access_keys, lambda row: row['access_key'],
        )


class KeyPairList(graphene.ObjectType):
    class Meta:
        interfaces = (PaginatedList, )

    items = graphene.List(KeyPair, required=True)


class KeyPairInput(graphene.InputObjectType):
    is_active = graphene.Boolean(required=False, default=True)
    is_admin = graphene.Boolean(required=False, default=False)
    resource_policy = graphene.String(required=True)
    concurrency_limit = graphene.Int(required=False)  # deprecated and ignored
    rate_limit = graphene.Int(required=True)

    # When creating, you MUST set all fields.
    # When modifying, set the field to "None" to skip setting the value.


class ModifyKeyPairInput(graphene.InputObjectType):
    is_active = graphene.Boolean(required=False)
    is_admin = graphene.Boolean(required=False)
    resource_policy = graphene.String(required=False)
    concurrency_limit = graphene.Int(required=False)  # deprecated and ignored
    rate_limit = graphene.Int(required=False)


class CreateKeyPair(graphene.Mutation):

    allowed_roles = (UserRole.SUPERADMIN,)

    class Arguments:
        user_id = graphene.String(required=True)
        props = KeyPairInput(required=True)

    ok = graphene.Boolean()
    msg = graphene.String()
    keypair = graphene.Field(lambda: KeyPair, required=False)

    @classmethod
    async def mutate(
        cls,
        root,
        info: graphene.ResolveInfo,
        user_id: uuid.UUID,
        props: KeyPairInput,
    ) -> CreateKeyPair:
        graph_ctx: GraphQueryContext = info.context
        # Check if user exists with requested email (user_id for legacy).
        from .user import users  # noqa
        query = (
            sa.select([users.c.uuid])
            .select_from(users)
            .where(users.c.email == user_id)
        )
        try:
            result = await graph_ctx.db_conn.execute(query)
            user_uuid = result.scalar()
            if user_uuid is None:
                return cls(ok=False, msg=f'User not found: {user_id}', keypair=None)
        except sa.exc.IntegrityError as e:
            return cls(ok=False, msg=f'integrity error: {e}', keypair=None)
        except (asyncio.CancelledError, asyncio.TimeoutError):
            raise
        except Exception as e:
            return cls(ok=False, msg=f'unexpected error: {e}', keypair=None)

        # Create keypair.
        ak, sk = generate_keypair()
        pubkey, privkey = generate_ssh_keypair()
        data = {
            'user_id': user_id,
            'access_key': ak,
            'secret_key': sk,
            'is_active': bool(props.is_active),
            'is_admin': bool(props.is_admin),
            'resource_policy': props.resource_policy,
            'concurrency_used': 0,
            'rate_limit': props.rate_limit,
            'num_queries': 0,
            'user': user_uuid,
            'ssh_public_key': pubkey,
            'ssh_private_key': privkey,
        }
        insert_query = (sa.insert(keypairs).values(data))
        try:
            result = await graph_ctx.db_conn.execute(insert_query)
            if result.rowcount > 0:
                # Read the created key data from DB.
                checkq = sa.select([keypairs]).where(keypairs.c.access_key == ak)
                result = await graph_ctx.db_conn.execute(checkq)
                o = KeyPair.from_row(info.context, result.first())
                return cls(ok=True, msg='success', keypair=o)
            else:
                return cls(ok=False, msg='failed to create keypair', keypair=None)
        except sa.exc.IntegrityError as e:
            return cls(ok=False, msg=f'integrity error: {e}', keypair=None)
        except (asyncio.CancelledError, asyncio.TimeoutError):
            raise
        except Exception as e:
            return cls(ok=False, msg=f'unexpected error: {e}', keypair=None)


class ModifyKeyPair(graphene.Mutation):

    allowed_roles = (UserRole.SUPERADMIN,)

    class Arguments:
        access_key = graphene.String(required=True)
        props = ModifyKeyPairInput(required=True)

    ok = graphene.Boolean()
    msg = graphene.String()

    @classmethod
    async def mutate(
        cls,
        root,
        info: graphene.ResolveInfo,
        access_key: AccessKey,
        props: ModifyUserInput,
    ) -> ModifyKeyPair:
        ctx: GraphQueryContext = info.context
        data: Dict[str, Any] = {}
        set_if_set(props, data, 'is_active')
        set_if_set(props, data, 'is_admin')
        set_if_set(props, data, 'resource_policy')
        set_if_set(props, data, 'rate_limit')
        update_query = (
            sa.update(keypairs)
            .values(data)
            .where(keypairs.c.access_key == access_key)
        )
        return await simple_db_mutate(cls, ctx, update_query)


class DeleteKeyPair(graphene.Mutation):

    allowed_roles = (UserRole.SUPERADMIN,)

    class Arguments:
        access_key = graphene.String(required=True)

    ok = graphene.Boolean()
    msg = graphene.String()

    @classmethod
    async def mutate(
        cls,
        root,
        info: graphene.ResolveInfo,
        access_key: AccessKey,
    ) -> DeleteKeyPair:
        ctx: GraphQueryContext = info.context
        delete_query = (
            sa.delete(keypairs)
            .where(keypairs.c.access_key == access_key)
        )
        return await simple_db_mutate(cls, ctx, delete_query)


class Dotfile(TypedDict):
    data: str
    path: str
    perm: str


def generate_keypair() -> Tuple[AccessKey, SecretKey]:
    '''
    AWS-like access key and secret key generation.
    '''
    ak = 'AKIA' + base64.b32encode(secrets.token_bytes(10)).decode('ascii')
    sk = secrets.token_urlsafe(30)
    return AccessKey(ak), SecretKey(sk)


def generate_ssh_keypair() -> Tuple[str, str]:
    '''
    Generate RSA keypair for SSH/SFTP connection.
    '''
    key = rsa.generate_private_key(
        backend=crypto_default_backend(),
        public_exponent=65537,
        key_size=2048
    )
    private_key = key.private_bytes(
        crypto_serialization.Encoding.PEM,
        crypto_serialization.PrivateFormat.TraditionalOpenSSL,
        crypto_serialization.NoEncryption()
    ).decode("utf-8")
    public_key = key.public_key().public_bytes(
        crypto_serialization.Encoding.OpenSSH,
        crypto_serialization.PublicFormat.OpenSSH
    ).decode("utf-8")
    return (public_key, private_key)


async def query_owned_dotfiles(
    conn: SAConnection,
    access_key: AccessKey,
) -> Tuple[List[Dotfile], int]:
    query = (
        sa.select([keypairs.c.dotfiles])
        .select_from(keypairs)
        .where(keypairs.c.access_key == access_key)
    )
    packed_dotfile = await conn.scalar(query)
    rows = msgpack.unpackb(packed_dotfile)
    return rows, MAXIMUM_DOTFILE_SIZE - len(packed_dotfile)


async def query_bootstrap_script(
    conn: SAConnection,
    access_key: AccessKey,
) -> Tuple[str, int]:
    query = (
        sa.select([keypairs.c.bootstrap_script])
        .select_from(keypairs)
        .where(keypairs.c.access_key == access_key)
    )
    script = await conn.scalar(query)
    return script, MAXIMUM_DOTFILE_SIZE - len(script)


def verify_dotfile_name(dotfile: str) -> bool:
    if dotfile in RESERVED_DOTFILES:
        return False
    return True



from __future__ import annotations

import attr
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ai.backend.common.logging import Logger

    from ..config import LocalConfig


@attr.s(auto_attribs=True, frozen=True)
class CLIContext:
    logger: Logger
    local_config: LocalConfig



from __future__ import annotations

import asyncio
from typing import (
    Callable,
    Final,
    TYPE_CHECKING,
)

from aioredis import Redis
from aioredlock import Aioredlock, LockError

if TYPE_CHECKING:
    from ai.backend.common.events import AbstractEvent, EventProducer


class GlobalTimer:

    """
    Executes the given async function only once in the given interval,
    uniquely among multiple manager instances across multiple nodes.
    """

    _lock_manager: Final[Aioredlock]
    _event_producer: Final[EventProducer]

    def __init__(
        self,
        redis: Redis,
        timer_name: str,
        event_producer: EventProducer,
        event_factory: Callable[[], AbstractEvent],
        interval: float = 10.0,
    ) -> None:
        self._lock_manager = Aioredlock([redis])
        self._event_producer = event_producer
        self._event_factory = event_factory
        self.lock_key = f"timer.{timer_name}.lock"
        self.interval = interval

    async def generate_tick(self) -> None:
        try:
            while True:
                try:
                    async with (
                        await self._lock_manager.lock(self.lock_key, lock_timeout=self.interval)
                    ) as lock:
                        await self._event_producer.produce_event(self._event_factory())
                        await self._lock_manager.extend(lock, lock_timeout=self.interval)
                        await asyncio.sleep(self.interval)
                except LockError:
                    pass
        except asyncio.CancelledError:
            pass
        finally:
            await self._lock_manager.destroy()

    async def join(self) -> None:
        self._tick_task = asyncio.create_task(self.generate_tick())

    async def leave(self) -> None:
        self._tick_task.cancel()
        await self._tick_task



"""add_shared_memory_to_resource_presets

Revision ID: c481d3dc6c7d
Revises: f5530eccf202
Create Date: 2020-04-20 14:10:35.591063

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = 'c481d3dc6c7d'
down_revision = 'f5530eccf202'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column('resource_presets', sa.Column('shared_memory', sa.BigInteger(), nullable=True))
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column('resource_presets', 'shared_memory')
    # ### end Alembic commands ###
